from .aaindex1 import *
# from .aaindex2 import *
# from .aaindex3 import *

#aaindex software metadata
__name__ = "aaindex"
__version__ = "1.0.5"
__description__ = "aaindex is a lightweight Python software package for accessing the data in the various AAIndex databases, which represent the physiochemical and biochemical properties of amino acids as numerical indices."
__author__ = 'AJ McKenna, https://github.com/amckenna41'
__authorEmail__ = 'amckenna41@qub.ac.uk'
__license__ = 'MIT'
__url__ = 'https://github.com/amckenna41/aaindex'
__download_url__ = "https://github.com/amckenna41/aaindex/archive/refs/heads/main.zip"
__status__ = "Production"
__maintainer__ = "AJ McKenna"
__keywords__ = ["amino acid index", "bioinformatics", "protein engineering", "python", "pypi", \
        "physiochemical properties", "biochemical properties", "proteins", "psp", "pysar"]
__test_suite__ = "tests"


#AAIndex2 

# class AAIndex2():

# get_ref_from_record()
# get_values_from_record

##AAINDEX 2 ###
###########
# class AAIndex2():

#     def __init__(self):
#         # Record.__init__(self)
#         self.key = None
#         self.desc = ''
#         self.ref = ''
#         self.authors = ''
#         self.title = ''
#         self.journal = ''
#         self.correlated = dict()
#         self.index = dict()
#         self.comment = ''
#         self.index = []
#         self.rows = dict()
#         self.cols = dict()
#         self._aaindex = dict()
#         self._aaindex = self._parse('data/aaindex2')

#     def extend(self, row):
#         self.index.append(row)

#     def _get(self, aai, aaj):
#         i = self.rows[aai]
#         j = self.cols[aaj]
#         return self.index[i][j]

#     def get(self, aai, aaj, d=None):
#         try:
#             return self._get(aai, aaj)
#         except:
#             pass
#         try:
#             return self._get(aaj, aai)
#         except:
#             return d

#     def __getitem__(self, aaij):
#         return self.get(aaij[0], aaij[1])

#     def median(self):
#         x = []
#         for y in self.index:
#             x.extend(filter(None, y))
#         x.sort()
#         if len(x) % 2 == 1:
#             return x[len(x) // 2]
#         return sum(x[len(x) // 2 - 1:len(x) // 2 + 1]) / 2.0

#     def _float_or_None(self, x):
#         if x == 'NA' or x == '-':
#             return None
#         return float(x)

#     def _parse(self, filename, quiet=True):
#         '''
#         Parse aaindex input file. `rec` must be `Record` for aaindex1 and
#         `MarixRecord` for aaindex2 and aaindex3.
#         '''
#         if not os.path.exists(filename):
#             if sys.version_info[0] < 3:
#                 from urllib import urlretrieve
#             else:
#                 from urllib.request import urlretrieve
#             url = 'ftp://ftp.genome.jp/pub/db/community/aaindex/' + os.path.split(filename)[1]
#             print('Downloading "%s"' % (url))
#             filename = urlretrieve(url, filename)[0]
#             print('Saved to "%s"' % (filename))
#         f = open(filename)

#         # current = rec()
#         lastkey = None

#         for line in f:
#             key = line[0:2]
#             if key[0] == ' ':
#                 key = lastkey

#             if key == '//':
#                 self._aaindex[self.key] = self
#                 self = rec()
#             elif key == 'H ':
#                 self.key = line[2:].strip()
#             elif key == 'R ':
#                 self.ref += line[2:]
#             elif key == 'D ':
#                 self.desc += line[2:]
#             elif key == 'A ':
#                 self.authors += line[2:]
#             elif key == 'T ':
#                 self.title += line[2:]
#             elif key == 'J ':
#                 self.journal += line[2:]
#             elif key == '* ':
#                 self.comment += line[2:]
#             elif key == 'C ':
#                 a = line[2:].split()
#                 for i in range(0, len(a), 2):
#                     self.correlated[a[i]] = float(a[i + 1])
#             # elif key == 'I ':
#             #     a = line[1:].split()
#             #     if a[0] != 'A/L':
#             #         current.extend(map(_float_or_None, a))
#             #     elif list(Record.aakeys) != [i[0] for i in a] + [i[-1] for i in a]:
#             #         print('Warning: wrong amino acid sequence for', current.key)
#             #     else:
#             #         try:
#             #             assert list(Record.aakeys[:10]) == [i[0] for i in a]
#             #             assert list(Record.aakeys[10:]) == [i[2] for i in a]
#             #         except:
#             #             print('Warning: wrong amino acid sequence for', current.key)
#             elif key == 'M ':
#                 a = line[2:].split()
#                 if a[0] == 'rows':
#                     if a[4] == 'rows':
#                         a.pop(4)
#                     assert a[3] == 'cols' and len(a) == 6
#                     i = 0
#                     for aa in a[2]:
#                         self.rows[aa] = i
#                         i += 1
#                     i = 0
#                     for aa in a[5]:
#                         self.cols[aa] = i
#                         i += 1
#                 else:
#                     self.extend(map(self._float_or_None, a))
#             elif not quiet:
#                 print('Warning: line starts with "%s"' % (key))

#             lastkey = key
        
#         return _aaindex

# aaindex2 = AAIndex2()

# # class Record:

# #     '''
# #     Amino acid index (AAindex) Record
# #     '''
# #     aakeys = 'ARNDCQEGHILKMFPSTWYV'

# #     def __init__(self):
# #         self.key = None
# #         self.desc = ''
# #         self.ref = ''
# #         self.authors = ''
# #         self.title = ''
# #         self.journal = ''
# #         self.correlated = dict()
# #         self.index = dict()
# #         self.comment = ''

# #     def extend(self, row):
# #         i = len(self.index)
# #         for x in row:
# #             self.index[self.aakeys[i]] = x
# #             i += 1

# #     def get(self, aai, aaj=None, d=None):
# #         assert aaj is None
# #         return self.index.get(aai, d)

# #     def __getitem__(self, aai):
# #         return self.get(aai)

# #     def median(self):
# #         x = sorted(filter(None, self.index.values()))
# #         half = len(x) // 2
# #         if len(x) % 2 == 1:
# #             return x[half]
# #         return (x[half - 1] + x[half]) / 2.0

# #     def __str__(self):
# #         desc = self.desc.replace('\n', ' ').strip()
# #         return '%s(%s: %s)' % (self.__class__.__name__, self.key, desc)


# # class MatrixRecord(Record):
# class MatrixRecord():

#     '''
#     Matrix record for mutation matrices or pair-wise contact potentials
#     '''

#     def __init__(self):
#         # Record.__init__(self)
#         self.key = None
#         self.desc = ''
#         self.ref = ''
#         self.authors = ''
#         self.title = ''
#         self.journal = ''
#         self.correlated = dict()
#         self.index = dict()
#         self.comment = ''
#         self.index = []
#         self.rows = dict()
#         self.cols = dict()

#     def extend(self, row):
#         self.index.append(row)

#     def _get(self, aai, aaj):
#         i = self.rows[aai]
#         j = self.cols[aaj]
#         return self.index[i][j]

#     def get(self, aai, aaj, d=None):
#         try:
#             return self._get(aai, aaj)
#         except:
#             pass
#         try:
#             return self._get(aaj, aai)
#         except:
#             return d

#     def __getitem__(self, aaij):
#         return self.get(aaij[0], aaij[1])

#     def median(self):
#         x = []
#         for y in self.index:
#             x.extend(filter(None, y))
#         x.sort()
#         if len(x) % 2 == 1:
#             return x[len(x) // 2]
#         return sum(x[len(x) // 2 - 1:len(x) // 2 + 1]) / 2.0

# _aaindex = dict()

# def _float_or_None(x):
#     if x == 'NA' or x == '-':
#         return None
#     return float(x)

# def _parse(filename, rec, quiet=True):
#     '''
#     Parse aaindex input file. `rec` must be `Record` for aaindex1 and
#     `MarixRecord` for aaindex2 and aaindex3.
#     '''
#     if not os.path.exists(filename):
#         if sys.version_info[0] < 3:
#             from urllib import urlretrieve
#         else:
#             from urllib.request import urlretrieve
#         url = 'ftp://ftp.genome.jp/pub/db/community/aaindex/' + os.path.split(filename)[1]
#         print('Downloading "%s"' % (url))
#         filename = urlretrieve(url, filename)[0]
#         print('Saved to "%s"' % (filename))
#     f = open(filename)

#     current = rec()
#     lastkey = None

#     for line in f:
#         key = line[0:2]
#         if key[0] == ' ':
#             key = lastkey

#         if key == '//':
#             _aaindex[current.key] = current
#             current = rec()
#         elif key == 'H ':
#             current.key = line[2:].strip()
#         elif key == 'R ':
#             current.ref += line[2:]
#         elif key == 'D ':
#             current.desc += line[2:]
#         elif key == 'A ':
#             current.authors += line[2:]
#         elif key == 'T ':
#             current.title += line[2:]
#         elif key == 'J ':
#             current.journal += line[2:]
#         elif key == '* ':
#             current.comment += line[2:]
#         elif key == 'C ':
#             a = line[2:].split()
#             for i in range(0, len(a), 2):
#                 current.correlated[a[i]] = float(a[i + 1])
#         # elif key == 'I ':
#         #     a = line[1:].split()
#         #     if a[0] != 'A/L':
#         #         current.extend(map(_float_or_None, a))
#         #     elif list(Record.aakeys) != [i[0] for i in a] + [i[-1] for i in a]:
#         #         print('Warning: wrong amino acid sequence for', current.key)
#         #     else:
#         #         try:
#         #             assert list(Record.aakeys[:10]) == [i[0] for i in a]
#         #             assert list(Record.aakeys[10:]) == [i[2] for i in a]
#         #         except:
#         #             print('Warning: wrong amino acid sequence for', current.key)
#         elif key == 'M ':
#             a = line[2:].split()
#             if a[0] == 'rows':
#                 if a[4] == 'rows':
#                     a.pop(4)
#                 assert a[3] == 'cols' and len(a) == 6
#                 i = 0
#                 for aa in a[2]:
#                     current.rows[aa] = i
#                     i += 1
#                 i = 0
#                 for aa in a[5]:
#                     current.cols[aa] = i
#                     i += 1
#             else:
#                 current.extend(map(_float_or_None, a))
#         elif not quiet:
#             print('Warning: line starts with "%s"' % (key))

#         lastkey = key

# # _parse('data/aaindex2', MatrixRecord)

# # print(_aaindex['KOLA920101'].journal)





################################################################################
#################                    AAIndex                   #################
################################################################################

#importing required modules and dependencies
import json
import numpy as np
import os
import sys, copy, re
import requests
import shutil
from difflib import get_close_matches
import csv
import urllib.request as request
from contextlib import closing

#global vars
DOWNLOAD_USING = 'ftp'

class AAIndex1():
    """
            Python parser for AAindex1: Amino Acid Index Database (AAI)
    The AAindex is a database of numerical indices representing various physiochemical
    and biochemical properties of amino acids and pairs of amino acids. This class 
    stores the amino acid index of 20 numerical values for the 20 amino acids 
    (http://www.genome.jp/aaindex/).
    Some of the parsing functionality is patrially inspired off the AAIndex1 parser
    by harmslab: https://github.com/harmslab/hops/blob/master/hops/features/data/util/aaindex2json.py

    Data format of AAI1:
    ************************************************************************
    *                                                                      *
    * Each entry has the following format.                                 *
    *                                                                      *
    * H Accession number                                                   *
    * D Data description                                                   *
    * R Pub med article ID (PMID)                                          *
    * A Author(s)                                                          *
    * T Title of the article                                               *
    * J Journal reference                                                  *
    * * Comment or missing                                                 *
    * C Accession numbers of similar entries with the correlation          *
    *   coefficients of 0.8 (-0.8) or more (less).                         *
    *   Notice: The correlation coefficient is calculated with zeros       *
    *   filled for missing values.                                         *
    * I Amino acid index data in the following order                       *
    *   Ala    Arg    Asn    Asp    Cys    Gln    Glu    Gly    His    Ile *
    *   Leu    Lys    Met    Phe    Pro    Ser    Thr    Trp    Tyr    Val *
    * //                                                                   *
    ************************************************************************
    --------------------------------------------------------------------------------
    H ANDN920101
    D alpha-CH chemical shifts (Andersen et al., 1992)
    R PMID:1575719
    A Andersen, N.H., Cao, B. and Chen, C.
    T Peptide/protein structure analysis using the chemical shift index method:
      upfield alpha-CH values reveal dynamic helices and aL sites
    J Biochem. and Biophys. Res. Comm. 184, 1008-1014 (1992)
    C BUNA790102    0.949
    I    A/L     R/K     N/M     D/F     C/P     Q/S     E/T     G/W     H/Y     I/V
        4.35    4.38    4.75    4.76    4.65    4.37    4.29    3.97    4.63    3.95
        4.17    4.36    4.52    4.66    4.44    4.50    4.35    4.70    4.60    3.95
    --------------------------------------------------------------------------------

    Parameters
    ----------
    None
    
    Methods
    -------
    parse_aaindex():
        parse AAI database into JSON format.
    parse_categories(aaindex_category_file):
        parse AAI indices into their respective categories.
    download_aaindex():
        download AAI database from its FTP server.
    amino_acids():
        get list of 20 amino acid letters.
    record_codes():
        get list of AAI record codes/Accession numbers.
    num_records():
        return total number of records in AAI database.
    record_names():
        return list of all descriptions for all records in AAI database.
    search():
        return 1 or more AAI records from its description.
    __getitem__():
        access full aaindex record using its record code.
    
    References
    ----------
    [1]: Nakai, K., Kidera, A., and Kanehisa, M.;  Cluster analysis of
         amino acid indices for prediction of protein structure and
         function.  Protein Eng. 2, 93-100 (1988)
    [2]: Tomii, K. and Kanehisa, M.;  Analysis of amino acid indices and
         mutation matrices for sequence comparison and structure
         prediction of proteins.  Protein Eng. 9, 27-36 (1996).
    [3]: Kawashima, S., Ogata, H., and Kanehisa, M.;  AAindex: amino acid
         index database.  Nucleic Acids Res. 27, 368-369 (1999).
    [4]: Kawashima, S. and Kanehisa, M.;  AAindex: amino acid index
         database.  Nucleic Acids Res. 28, 374 (2000).
    """
    def __init__(self):

        self.aaindex_module_path = os.path.dirname(os.path.abspath(sys.modules[self.__module__].__file__))
        self.data_dir = "data"
        self.aaindex_filename = "aaindex1"
        self.download_using = DOWNLOAD_USING

        #download AAI database using ftp or https
        if self.download_using =='ftp':
            self.url = "ftp://ftp.genome.jp/pub/db/community/aaindex/aaindex1"
        elif self.download_using =='http' or self.download_using =='https':
            self.url = "https://www.genome.jp/ftp/db/community/aaindex/aaindex1"
        else:
            self.url = "ftp://ftp.genome.jp/pub/db/community/aaindex/aaindex1"

        #if AAIndex database not found in data directory then call download method
        if not (os.path.isfile(os.path.join(self.aaindex_module_path, self.data_dir, self.aaindex_filename))):
            self.download_aaindex()

        #get dict of categories
        self.categories = self.get_all_categories()

        #if parsed json of AAIndex already in file then read it and return 
        if (os.path.isfile(os.path.join(self.aaindex_module_path, self.data_dir, self.aaindex_filename + '.json'))):
            with open(os.path.join(self.aaindex_module_path, self.data_dir, self.aaindex_filename + '.json')) as aai_json:
                self.aaindex_json = json.load(aai_json)
        else:
            #parse AAIndex database file into JSON format
            self.aaindex_json = self.parse_aaindex()

        #get last updated - as shown on homepage (https://www.genome.jp/aaindex/)
        self.last_updated = "February 13, 2017" 

    def parse_aaindex(self):
        """
        Parse AAI database into JSON format. Each AAI record will be indexed by
        its accession number/index code, and will be in the format as shown in the
        docstring above, with the addition of a category key which shows the type/
        category of the record in question. The file will be stored in a json file 
        determined by AAINDEX_FILENAME variable.

        Returns
        -------
        :aaindex_json : dict
          parsed AAI database in dict form.
        """
        #initialise keys of AAI database
        template_dict = {"H":[],
                         "D":[],
                         "R":[],
                         "A":[],
                         "*":[],
                         "T":[],
                         "J":[],
                         "C":[],
                         "I":[]}

        #open AAI file for reading and parsing, by default it should be stored in self.data_dir
        try:
            tmp_filepath = os.path.join(self.aaindex_module_path, self.data_dir, self.aaindex_filename)
            f = open(tmp_filepath,'r')
        except IOError:
            print('Error opening file, check filename = {} and is stored in {} directory.'.format(
                    self.aaindex_filename, self.data_dir))

        #read lines of file
        lines = f.readlines()
        f.close()

        #pattern that seperates each AAI record
        clean_up_pattern = re.compile("\"")

        #initilaise parsed AAI database dictionary
        aaindex_json = {}
        current_dict = copy.deepcopy(template_dict)

        '''
        iterate through each line in the AAI database, parsing each record and its
        amino acid values into its own entry in the dictionary. Each index/record
        is seperated by a '//'. Remove any duplicate records, set any missing
        ('-') or NA amino acid values to 0. Store resulting dict into aaindex_json
        instance variable.
        '''
        for l in lines:
            if l.startswith("//"):

                # handle meta data of each record - name, desc, citation, notes
                name = " ".join(current_dict["H"])
                name = clean_up_pattern.sub("'", name)

                description = " ".join(current_dict["D"])
                description = clean_up_pattern.sub("'", description)

                #append author, title and journal name to reference
                references = "{} '{}' {}".format(" ".join(current_dict["A"]),
                                               " ".join(current_dict["T"]),
                                               " ".join(current_dict["J"]))
                references = clean_up_pattern.sub("'", references)

                # references = references + "; Kawashima, S. and Kanehisa, M."
                #     "AAindex: amino acid index database.'  Nucleic Acids Res. 28, 374 (2000)."

                #parse pub med article ID
                pmid = " ".join(current_dict["R"])
                pmid = clean_up_pattern.sub("'", pmid)
                pmid = pmid.replace("PMID:", "")

                #appending correlation coefficient of current record, convert into dict
                correlation_coefficient = " ".join(current_dict["C"])
                correlation_coefficient = clean_up_pattern.sub("'", correlation_coefficient)
                correlation_coefficient_ = {}
                correlation_coefficient_list = [correlation_coefficient.split()[n:n+2] for n in range(0, len(correlation_coefficient.split()), 2)]

                for correlation in correlation_coefficient_list:
                    correlation_coefficient_[correlation[0]] = correlation[1]
            
                #parse notes from record
                notes = " ".join(current_dict["*"])
                notes = clean_up_pattern.sub("'", notes)

                # parse individual amino acid values
                aa_lines = current_dict["I"]

                aa_names = aa_lines[0].split()
                row_0_names = [aa.split("/")[0] for aa in aa_names]
                row_1_names = [aa.split("/")[1] for aa in aa_names]

                row_0_values = aa_lines[1].split()
                row_1_values = aa_lines[2].split()

                #get values for each row of amino acids in record
                values = {}
                for i in range(len(row_0_values)):
                    try:
                        values[row_0_names[i]] = float(row_0_values[i])
                    except ValueError:
                        values[row_0_names[i]] = "NA"

                    try:
                        values[row_1_names[i]] = float(row_1_values[i])
                    except ValueError:
                        values[row_1_names[i]] = "NA"

                #look for duplicate name entries
                try:
                    aaindex_json[name]
                    err = "duplicate value name ({})".format(name)
                    raise ValueError('Duplicate AAI Record name found')
                except KeyError:
                    pass

                #append all aaindex record data to object
                aaindex_json[name] = {"description": description,
                                  "references": references,
                                  "pmid": pmid,
                                  "correlation_coefficients": correlation_coefficient_,
                                  "notes": notes,
                                  "values": values
                            }
                
                #create copy of dict 
                current_dict = copy.deepcopy(template_dict)
                continue

            this_entry = l[0]
            if l[0] != " ":
                current_entry = this_entry

            current_dict[current_entry].append(l[1:].strip())
        
        #append '-' to each aa index entry to account for missing AA in a protein sequence
        for index in aaindex_json:

            #add category to aaindex file
            aaindex_json[index]['category'] = self.categories[index]
            aaindex_json[index]['values']['-'] = 0

            #set any NA amino acid values to 0
            for val in aaindex_json[index]['values']:
                if aaindex_json[index]['values'][val] == 'NA':
                    aaindex_json[index]['values'][val] = 0

        #save parsed dictionary into JSON format to self.data_dir
        with open((os.path.join(self.aaindex_module_path, self.data_dir, 
            self.aaindex_filename + '.json')),'w') as output_F:
          json.dump(aaindex_json, output_F, indent=4, sort_keys=True)

        return aaindex_json

    def download_aaindex(self, save_dir="data"):
        """
        If AAI database not found in DATA directory, then it will be downloaded from
        the dedicated FTP or HTTPS server from https://www.genome.jp/aaindex/.
        FTP is the default method used for downloading the database with the URL:
        "ftp://ftp.genome.jp/pub/db/community/aaindex/aaindex1". If you want to
        download by https, set the 'DOWNLOAD_USING' variable to 'https'. The class
        looks for the AAI database and its parsed JSON form in the 'data' directory, an
        error will throw if not found.

        Parameters
        ----------
        :save_dir : str (default = 'data')
            Directory to save the AAI database to. By default it wil be stored in
            the 'data' dir. Parameter has to point to a directory within the aaindex 
            directory or there will be an error on instantiation of the class. 
        """
        #fetch AAI database from URL if not present in 'data' dir
        try:
            if not(os.path.isfile(os.path.join(self.aaindex_module_path, save_dir, self.aaindex_filename))):
                try:
                    with closing(request.urlopen(self.url)) as r:
                        with open((os.path.join(self.aaindex_module_path, save_dir, self.aaindex_filename)), 'wb') as f:
                            shutil.copyfileobj(r, f)
                    print('AAIndex1 successfully downloaded.')
                except requests.exceptions.RequestException:
                    print('Error downloading or exporting AAIndex from the url {}.'.format(self.url))
            else:
                pass    #AAIndex already in folder
        except:
            raise OSError('Save Directory does not exist: {}.'.format(save_dir))

    def parse_categories(self, aaindex_category_file='aaindex_to_category.txt'):
        """
        Parse category file which maps each AAI record in the databse into 1 of 8 categories.
        Category file and parsing code inspired from: https://github.com/harmslab/hops. 

        Parameters
        ----------
        :aaindex_category_file : str
            Name of category file to parse (default is "aaindex_to_category.txt"). If parameter
            is a file name then it will be imported from the 'data' directory by default, else
            it will be imported from its full path.

        Returns
        -------
        :aaindex_category : dict
            Dictionary that maps each AAI record into 1 of 8 categories.
        """
        #if input parameter is a full path, read it else read from default 'data'
        if os.path.isfile(aaindex_category_file):
            try:
                f = open(aaindex_category_file,'r')
            except IOError:
                print('Error opening AAIndex1 category file {}.'
                    .format(aaindex_category_file))
        else:
            try:
                f = open((os.path.join(self.aaindex_module_path, self.data_dir, aaindex_category_file)),'r')
            except IOError:
                print('Error opening AAIndex1 category file, check {} is in {} directory.'
                    .format(aaindex_category_file, self.data_dir))

        #get total number of lines in file
        # total_lines = len(f.readlines(  ))

        #open new file in data directory to store parsed category file
        f.seek(0)
        category_output_file = "aaindex_categories.txt"
        f_out = open((os.path.join(self.aaindex_module_path, self.data_dir, category_output_file)), "w")

        #lines starting with '#' are file metadata so don't write these to parsed output file
        for line in f.readlines():
            if not (line.startswith('#')):
                f_out.write(line)

        #close both files
        f.close()
        f_out.close()

        #open parsed category file for reading
        with open(os.path.join(self.aaindex_module_path, self.data_dir, category_output_file)) as f_out:
            reader = csv.reader(f_out, delimiter="\t")
            d = list(reader)

        aaindex_category = {}

        #iterate through all lines in parsed file and store AAI indices as keys
        #  in the output dict and their respective categories as the values of dict
        for i in range(0, len(d)):
          for j in range(0, len(d[i])):
            category_substring = d[i][1].strip()
            category_substring = category_substring.split(" ", 1)
            aaindex_category[d[i][0]] = category_substring[0]
        
        for code, cat in aaindex_category.items():
            self.aaindex_json[code]["category"] = cat

        return aaindex_category

    def get_all_categories(self, category_file="aaindex_categories.txt"):
        """
        Return list of all indices, their description and associated categories from 
        parsed aaindex_categories.txt file created from parse_categories() function.

        Parameters
        ----------
        :category_file : str (default = "aaindex_categories.txt")
            path to categories mapping file.
        
        Returns
        -------
        :aaindex_category : dict
            Dictionary that maps each AAI record into 1 of 8 categories.
        """
        #if parsed categories file doesn't exist in 'data' then call function to get it
        if not (os.path.isfile(os.path.join(self.aaindex_module_path, self.data_dir, category_file))):
            self.parse_categories()

        #read categories file and its content
        try:
            with open(os.path.join(self.aaindex_module_path, self.data_dir, category_file)) as f_out:
                reader = csv.reader(f_out, delimiter="\t")
                d = list(reader)
        except IOError:
            print('Error opening AAIndex1 category file {}.'.format(category_file))

        aaindex_category = {}

        #iterate through all lines in parsed file and store AAI indices as keys
        #   in the output dict and their respective categories as the values of dict
        for i in range(0, len(d)):
          for j in range(0, len(d[i])):
            category_substring = d[i][1].strip()
            category_substring = category_substring.split(" ", 1)
            aaindex_category[d[i][0]] = category_substring[0]

        return aaindex_category

    def search(self, description):
        """
        Return 1 or more full AAI database records from their description. Search
        through the descriptions/names of all records in the database, returning
        record(s) that match description input parameter.

        Parameters
        ----------
        :description : str
            keywords to search for in all AAI1 records.

        Returns
        -------
        :all_indices : dict
            AAIndex record(s) that contain the keywords in the description input parameter.
        """
        all_indices = {}

        #if input param not a list or string then raisse type error
        if not (isinstance(description, list)) and not (isinstance(description, str)):
            raise TypeError("Input Description parameter must be a list or str, got {}.".format(type(description)))

        #convert description parameter to list to make iterable
        if not (isinstance(description, list)):
            description = [description]

        #iterate over description list, if keywords are in an aaindex record's description
        #   then add record to all_indices object
        for desc in description:
            for index, value in self.aaindex_json.items():
                if (desc.lower() in self.aaindex_json[index]['description'].lower()):
                    all_indices[index] = self.aaindex_json[index]

        return all_indices

    def amino_acids(self):
        """
        Get all canonical amino acid letters. The '-' value will also be included
        in the list from this function as it accounts for the abcense of any
        amino acid or gaps in a AAI record.

        Returns
        -------
        :amino_acids : list
            List of all 20 canonical amino acid letters as found in each record
            of the AAI database.
        """
        amino_acids = list(self.aaindex_json[list(self.aaindex_json.keys())[0]]["values"].keys())
        amino_acids.sort()  #sort into alphabetical order
        
        return amino_acids

    def record_codes(self):
        """
        Get list of all AAI index codes/Accession numbers for each record in the database.

        Returns
        -------
        :records : list
            list of record names/index codes for all records in the AAI database.
        """
        records = list(self.aaindex_json.keys())
        records.sort()  #sort into alphabetical order

        return records

    def num_records(self):
        """
        Calculate total number of records/indices in the AAI database.

        Returns
        -------
        :len(aaindex.record_codes()) : int
            number of indices/records found in the AAI database.
        """
        return len(self.record_codes())

    def record_names(self):
        """
        Return a list of all index descriptions for all records in the AAI database.

        Returns
        -------
        :desc : list
            list of descriptions for all records in the AAI database.
        """
        desc = []

        #iterate through database, appending all descriptions to desc list
        for name in list(self.aaindex_json.values()):
            desc.append(name['description'])

        return desc

######################          Getters & Setters          ######################

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, value):
        self._url = value

    @property
    def categories(self):
        return self._categories

    @categories.setter
    def categories(self, value):
        self._categories = value

    @property
    def last_updated(self):
        return self._last_updated

    @last_updated.setter
    def last_updated(self, value):
        self._last_updated = value

################################################################################

    # def __repr__(self):
    #     return "<AAIndex: {}>".format(self)

    def __sizeof__(self):
        """ Return size of AAI database file """
        return os.path.getsize(os.path.isfile(os.path.join(self.data_dir, self.aaindex_filename)))

    def __getitem__(self, record_code):
        """
        Return full AAI database record details from its index code/Accession number by
        making the whole database subscriptable.

        Parameters
        ----------
        :record_code : str
            AAI database record Accession number.

        Returns
        -------
        :record : dict
            dict of AAI database record.

        Usage
        -----
            aaindex = AAIndex()
            full_record = aaindex['ZIMJ680105']
        """
        #stripping input of whitespace
        try:
            record_code = record_code.strip()
        except:
            raise TypeError('Input parameter {} is not of correct datatype string, got {}' \
                .format(record_code, type(record_code)))

        #check that inputted record_code/Accession num does exist in the AAI database
        if record_code not in (self.record_codes()):
            raise ValueError('Record Index ({}) not found in AAIndex'.format(record_code))
        
        #get full record from parsed JSON
        record = self.aaindex_json[record_code]

        #create instance of map class which allows you to access dict using dot notation 
        record = Map(record)

        return record

class Map(dict):
    """
    Instantiating this class will convert a dict such that it can be accessed using 
    dot notation which makes it easier for accessing the individual elements of the 
    aaindex records. It also works for nested dicts.

    Parameters 
    ----------
    :dict : dict 
        input dictionary to be mapped into dot notation.

    Usage
    -----
    m = Map({'first_name': 'Eduardo'}, last_name='Pool', age=24, sports=['Soccer'])
    # Add new key
    m.new_key = 'Hello world!'
    # Or
    m['new_key'] = 'Hello world!'
    print m.new_key
    print m['new_key']
    # Update values
    m.new_key = 'Yay!'
    # Or
    m['new_key'] = 'Yay!'
    # Delete key
    del m.new_key
    # Or
    del m['new_key']
    
    Reference
    ---------
    [1]: https://stackoverflow.com/questions/2352181/how-to-use-a-dot-to-access-members-of-dictionary
    """
    def __init__(self, *args, **kwargs):
        super(Map, self).__init__(*args, **kwargs)
        for arg in args:
            if isinstance(arg, dict):
                for k, v in arg.items():
                    self[k] = v

        if kwargs:
            for k, v in kwargs.items():
                self[k] = v

    def __getattr__(self, attr):
        return self.get(attr)

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __setitem__(self, key, value):
        super(Map, self).__setitem__(key, value)
        self.__dict__.update({key: value})

    def __delattr__(self, item):
        self.__delitem__(item)

    def __delitem__(self, key):
        super(Map, self).__delitem__(key)
        del self.__dict__[key]

#create instance of aaindex1 class
aaindex1 = AAIndex1()


###############################################################################
#####   Setup.py - installs all the required packages and dependancies    #####
###############################################################################

import pathlib
from setuptools import setup, find_packages
import sys
import aaindex 

#ensure python version is greater than 3
if sys.version_info[0] < 3:
    sys.exit('Python 3 is the minimum version requirement')

#parse readme file
HERE = pathlib.Path(__file__).parent
README = (HERE / 'README.md').read_text()

setup(name=aaindex.__name__,
      version=aaindex.__version__,
      description=aaindex.__description__,
      long_description = README,
      long_description_content_type = "text/markdown",
      author=aaindex.__author__,
      author_email=aaindex.__authorEmail__,
      maintainer=aaindex.__maintainer__,
      license=aaindex.__license__,
      url=aaindex.__url__,
      download_url=aaindex.__download_url__,
      keywords=aaindex.__keywords__,
      classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Healthcare Industry',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
      install_requires=[
          'numpy>=1.16.6',
        #   'delayed',
          'requests>=2.25',
          'urllib3>=1.26'
      ],

     packages=find_packages(),
     include_package_data=True,
     zip_safe=True)


################################################################################
#################             AAIndex Module Tests             #################
################################################################################

import os
import numpy as np
import unittest
import requests
from importlib.metadata import metadata
from aaindex import aaindex1, __version__

class AAIndexTests(unittest.TestCase):

    def setUp(self):
        """ Inititalise AAIndex module variables and test data directory. """
        self.test_dir = 'test_data'

        #make test data directory, remove old one if exists
        if (os.path.isfile(self.test_dir)):
            os.rmdir(self.test_dir)
        os.mkdir(self.test_dir)

    def test_aaindex_metadata(self):
        """ Testing correct aaindex version and metadata. """
        self.assertEqual(__version__, "1.0.5",
            "aaindex version is not correct, got: {}".format(metadata("aaindex")['version']))
        self.assertEqual(metadata("aaindex")['name'], "aaindex", 
            "aaindex software name is not correct, got: {}".format(metadata("aaindex")['name']))
        self.assertEqual(metadata("aaindex")['author'], "AJ McKenna, https://github.com/amckenna41", 
            "aaindex author is not correct, got: {}".format(metadata("aaindex")['author']))
        self.assertEqual(metadata("aaindex")['author-email'], "amckenna41@qub.ac.uk", 
            "aaindex author email is not correct, got: {}".format(metadata("aaindex")['author-email']))
        self.assertEqual(metadata('aaindex')['home-page'], "https://github.com/amckenna41/aaindex", 
            "aaindex home page url is not correct, got: {}".format(metadata('aaindex')['home-page']))
        self.assertEqual(metadata('aaindex')['maintainer'], "AJ McKenna", 
            "aaindex maintainer is not correct, got: {}".format(metadata('aaindex')['maintainer']))
        self.assertEqual(metadata('aaindex')['license'], "MIT", 
            "aaindex license type is not correct, got: {}".format(metadata('aaindex')['license']))
        self.assertEqual(metadata('aaindex')['summary'], 
            "aaindex is a lightweight Python software package for accessing the data in the various AAIndex databases,"
            " which represent the physiochemical and biochemical properties of amino acids as numerical indices.", 
                    "aaindex package summary is not correct, got: {}".format(metadata('aaindex')['summary']))
        self.assertEqual(metadata('aaindex')['keywords'], 
            "amino acid index,bioinformatics,protein engineering,python,pypi,physiochemical properties,"
            "biochemical properties,proteins,psp,pysar", 
                "aaindex keywords are not correct, got: {}".format(metadata('aaindex')['keywords']))

    @unittest.skip("Don't want to overload the FTP server each time tests are run.")
    def test_download(self):
        """ Test Case to check that the download functionality works for the required AAIndex files.
        The file will firstly be removed from the test directory and then redownloaded, and
        its presence in the directory will pass the test. """
#1.)
        #if AAI1 present in test dir then remove
        if (os.path.isfile(os.path.join(self.TEST_DIR, 'test_aaindex1'))):
            os.remove(os.path.join(self.TEST_DIR, 'test_aaindex1'))

        #assert that OSError exception raised when erroneous directory input to download func
        with self.assertRaises(OSError):
            aaindex1.download_aaindex()
#2.)
        #download AAI1 to local test directory
        aaindex1.download_aaindex()

        #verify download functionality has worked as desired
        self.assertTrue(os.path.isfile(os.path.join(self.TEST_DIR, 'aaindex1')),
            'AAI1 did not download correctly to the {} directory.'.format(self.TEST_DIR))

    @unittest.skip("Don't want to overload the FTP server each time tests are run.")
    def test_url(self):
        """ Test Case to check that the URL endpoints used for downloading the AAI1 databases
        return a 200 status code. """

        AA_INDEX1_URL_FTP = "https://www.genome.jp/ftp/db/community/aaindex/aaindex1"
        AA_INDEX1_URL_HTTPS = "https://www.genome.jp/ftp/db/community/aaindex/aaindex1"
        AA_INDEX2_URL = "https://www.genome.jp/ftp/db/community/aaindex/aaindex2"
        AA_INDEX3_URL = "https://www.genome.jp/ftp/db/community/aaindex/aaindex3"
#1.)
        #test URL endpoints for AAINDEX are active and give a 200 status code
        r = requests.get(AA_INDEX1_URL_FTP, allow_redirects = True)
        self.assertEqual(r.status_code, 200, 'URL not returning Status Code 200.')

        r = requests.get(AA_INDEX1_URL_HTTPS, allow_redirects = True)
        self.assertEqual(r.status_code, 200, 'URL not returning Status Code 200.')

        r = requests.get(AA_INDEX2_URL, allow_redirects = True)
        self.assertEqual(r.status_code, 200, 'URL not returning Status Code 200.')

        r = requests.get(AA_INDEX3_URL, allow_redirects = True)
        self.assertEqual(r.status_code, 200, 'URL not returning Status Code 200.')

    def test_amino_acids(self):
        """ Test Case to check that only valid Amino Acids are used within the AAIndex class. """
        valid_amino_acids = ['-', 'A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K',
            'L', 'M', 'N', 'P', 'Q', 'R', 'S', 'T', 'V', 'W', 'Y']
        amino_acids = aaindex1.amino_acids()
#1.)
        for aa in amino_acids:
            self.assertIn(aa, valid_amino_acids, 'Amino Acid {} not in valid \
                amino acids: {}.'.format(aa, valid_amino_acids))

    def test_num_records(self):
        """ Test Case to check that the correct number of indices/records are present
        in the AAIndex object. To date, 566 indices are present in the database. The 
        test takes into account if more features are added to the database in the future. """
#1.)
        self.assertTrue(aaindex1.num_records() >= 566, 'Incorrect number \
            of features found in AAI1, wanted 566, got {}.'.format(aaindex1.num_records()))

    def test_feature_size(self):
        """ Test Case to check that the AAI1 has the correct dimensionality. Check that
        the number of keys in parsed JSON from the AAI1 is correct. The test takes
        into account if more features are added to the database in the future. """
#1.)
        self.assertTrue(len(aaindex1.record_codes()) >= 566, 'Incorrect number \
            of features found in AAI1, wanted 566, got {}.'.format(len(aaindex1.record_codes())))

    def test_records(self):
        """ Test Case to check that the correct record data and amino acid values 
        are returned for the accessiom number/index codes using the 
        __getitem__ function. """
#1.)
        #initialise test feature codes and their correct amino acid values
        index_code1 = 'AURR980103'
        expected_index_code1_vals = {'A': 1.05, 'L': 0.96, 'R': 0.81, 'K': 0.97, 'N': 0.91,
            'M': 0.99, 'D': 1.39, 'F': 0.95, 'C': 0.6, 'P': 1.05, 'Q': 0.87,
            'S': 0.96, 'E': 1.11, 'T': 1.03, 'G': 1.26, 'W': 1.06, 'H': 1.43,
            'Y': 0.94, 'I': 0.95, 'V': 0.62, '-': 0}

        index_code2 = 'FINA770101'
        expected_index_code2_vals = {'-': 0, 'A': 1.08, 'C': 0.95, 'D': 0.85, 'E': 1.15, 
            'F': 1.1, 'G': 0.55, 'H': 1.0, 'I': 1.05, 'K': 1.15, 'L': 1.25, 'M': 1.15, 
            'N': 0.85, 'P': 0.71, 'Q': 0.95, 'R': 1.05, 'S': 0.75, 'T': 0.75, 
            'V': 0.95, 'W': 1.1, 'Y': 1.1}

        index_code3 = 'NAGK730101'
        expected_index_code3_vals = {'A': 1.29, 'L': 1.23, 'R': 0.83, 'K': 1.23, 'N': 0.77,
            'M': 1.23, 'D': 1.0, 'F': 1.23, 'C': 0.94, 'P': 0.7, 'Q': 1.1,
            'S': 0.78, 'E': 1.54, 'T': 0.87, 'G': 0.72, 'W': 1.06, 'H': 1.29,
            'Y': 0.63, 'I': 0.94, 'V': 0.97, '-': 0}

        index_code4 = 'TANS770103'
        expected_index_code4_vals = {'-': 0, 'A': 0.79, 'C': 1.268, 'D': 0.53, 'E': 0.643, 
            'F': 1.052, 'G': 0.725, 'H': 0.864, 'I': 1.361, 'K': 0.735, 'L': 1.111, 
            'M': 1.092, 'N': 0.832, 'P': 1.249, 'Q': 1.038, 'R': 1.087, 'S': 1.093, 
            'T': 1.214, 'V': 1.428, 'W': 1.114, 'Y': 1.34}

        index_code5 = "ABCDEFGH"
        index_code6 = "123456"
        index_code7 = "blahblahblah"
#1.)
        #get amino acid values for inputted feature/index codes
        record = aaindex1[index_code1]

        self.assertIn('description', list(record.keys()), 
            "description key not found in aaindex record object.")
        self.assertIn('references', list(record.keys()), 
            "reference key not found in aaindex record object.")
        self.assertIn('notes', list(record.keys()), 
            "notes key not found in aaindex record object.")
        self.assertIn('values', list(record.keys()), 
            "values key not found in aaindex record object.")
        self.assertIn('category', list(record.keys()), 
            "category key not found in aaindex record object.")
        self.assertIn('pmid', list(record.keys()), 
            "pmid key not found in aaindex record object.")
        self.assertIn('correlation_coefficients', list(record.keys()), 
            "correlation_coefficients key not found in aaindex record object.")

        self.assertEqual(record['description'], "Normalized positional residue frequency at helix termini N' (Aurora-Rose, 1998)",
            "Description from AA1 record do not match the expected value.")
        self.assertEqual(record['notes'], '',
            "Notes from AA1 record do not match the expected value.")
        self.assertTrue(record['references'].startswith("Aurora, R. and Rose, G."),
            "Reference from AA1 record do not match the expected value.")
        self.assertEqual(record['category'], 'sec_struct', 
            "Category from AA1 record do not match the expected value.")
        self.assertEqual(record['pmid'], '9514257', 
            "PMID from AA1 record do not match the expected value.")
        self.assertEqual(record['correlation_coefficients'], {}, 
            "Correlation Coefficients from AA1 record do not match the expected value.")
        self.assertEqual(record['values'], expected_index_code1_vals, 
            "Values for individual amino acids do not match expected.")
        for aa, val in list(record['values'].items()):
            if (aa == '-'):
                continue
            self.assertIsInstance(val, float, 
                "Values in record values not of correct datatype, expected float, got {}.".format(type(val)))
#2.)
        record = aaindex1[index_code2]

        self.assertIn('description', list(record.keys()), 
            "description key not found in aaindex record object.")
        self.assertIn('references', list(record.keys()), 
            "reference key not found in aaindex record object.")
        self.assertIn('notes', list(record.keys()), 
            "notes key not found in aaindex record object.")
        self.assertIn('values', list(record.keys()), 
            "values key not found in aaindex record object.")
        self.assertIn('category', list(record.keys()), 
            "category key not found in aaindex record object.")
        self.assertIn('pmid', list(record.keys()), 
            "pmid key not found in aaindex record object.")
        self.assertIn('correlation_coefficients', list(record.keys()), 
            "correlation_coefficients key not found in aaindex record object.")

        self.assertEqual(record['description'], 'Helix-coil equilibrium constant (Finkelstein-Ptitsyn, 1977)',
            "Description from AA1 record do not match the expected value.")
        self.assertEqual(record['notes'], '', 
            "Notes from AA1 record do not match the expected value.")
        self.assertTrue(record['references'].startswith("Finkelstein, A.V. and Ptitsyn"), 
            "Reference from AA1 record do not match the expected value.")
        self.assertEqual(record['category'], 'sec_struct', 
            "Category from AA1 record do not match the expected value.")
        self.assertEqual(record['pmid'], '843599', 
            "PMID from AA1 record do not match the expected value.")
        self.assertEqual(record['correlation_coefficients'], {'AURR980109': '0.802', 
            'AURR980113': '0.849', 'AURR980114': '0.875', 'KANM800103': '0.823', 
            'MAXF760101': '0.810', 'PTIO830101': '0.826', 'QIAN880106': '0.810', 
            'QIAN880107': '0.814', 'SUEM840101': '0.883'}, 
            "Correlation Coefficients from AA1 record do not match the expected value.")
        self.assertEqual(record['values'], expected_index_code2_vals, 
            "Values for individual amino acids do not match expected.")
        for aa, val in list(record['values'].items()):
            if (aa == '-'):
                continue
            self.assertIsInstance(val, float, 
                "Values in record values not of correct datatype, expected float, got {}.".format(type(val)))
#3.)
        record = aaindex1[index_code3]

        self.assertIn('description', list(record.keys()), 
            "description key not found in aaindex record object.")
        self.assertIn('references', list(record.keys()), 
            "reference key not found in aaindex record object.")
        self.assertIn('notes', list(record.keys()), 
            "notes key not found in aaindex record object.")
        self.assertIn('values', list(record.keys()), 
            "values key not found in aaindex record object.")
        self.assertIn('category', list(record.keys()), 
            "category key not found in aaindex record object.")
        self.assertIn('pmid', list(record.keys()), 
            "pmid key not found in aaindex record object.")
        self.assertIn('correlation_coefficients', list(record.keys()), 
            "correlation_coefficients key not found in aaindex record object.")

        self.assertEqual(record['description'], 'Normalized frequency of alpha-helix (Nagano, 1973)',
            "Description from AA1 record do not match the expected value.")
        self.assertEqual(record['notes'], '',
            "Notes from AA1 record do not match the expected value.")
        self.assertTrue(record['references'].startswith("Nagano, K."),
            "Reference from AA1 record do not match the expected value.")
        self.assertEqual(record['category'], 'sec_struct', 
            "Category from AA1 record do not match the expected value.")
        self.assertEqual(record['pmid'], '4728695', 
            "PMID from AA1 record do not match the expected value.")
        self.assertEqual(record['correlation_coefficients'], 
            {'BURA740101': '0.883', 'CHOP780201': '0.886', 'CRAJ730101': '0.925', 
            'GEIM800101': '0.912', 'GEIM800104': '0.828', 'ISOY800101': '0.862', 
            'KANM800101': '0.883', 'LEVM780101': '0.894', 'LEVM780104': '0.918', 
            'MAXF760101': '0.877', 'NAGK730103': '-0.870', 'PALJ810101': '0.953', 
            'PALJ810102': '0.876', 'PRAM900102': '0.894', 'RACS820108': '0.820', 
            'ROBB760101': '0.910', 'TANS770101': '0.925'}, 
            "Correlation Coefficients from AA1 record do not match the expected value.")
        self.assertEqual(record['values'], expected_index_code3_vals, 
            "Values for individual amino acids do not match expected.")
        for aa, val in list(record['values'].items()):
            if (aa == '-'):
                continue
            self.assertIsInstance(val, float, 
                "Values in record values not of correct datatype, expected float, got {}.".format(type(val)))
#4.)
        record = aaindex1[index_code4]

        self.assertIn('description', list(record.keys()), 
            "description key not found in aaindex record object.")
        self.assertIn('references', list(record.keys()), 
            "reference key not found in aaindex record object.")
        self.assertIn('notes', list(record.keys()), 
            "notes key not found in aaindex record object.")
        self.assertIn('values', list(record.keys()), 
            "values key not found in aaindex record object.")
        self.assertIn('category', list(record.keys()), 
            "category key not found in aaindex record object.")
        self.assertIn('pmid', list(record.keys()), 
            "pmid key not found in aaindex record object.")
        self.assertIn('correlation_coefficients', list(record.keys()), 
            "correlation_coefficients key not found in aaindex record object.")

        self.assertEqual(record['description'], 'Normalized frequency of extended structure (Tanaka-Scheraga, 1977)',
            "Description from AA1 record do not match the expected value.")
        self.assertEqual(record['notes'], '', 
            "Notes from AA1 record do not match the expected value.")
        self.assertTrue(record['references'].startswith("Tanaka, S. and Scheraga, H.A."),
            "Reference from AA1 record do not match the expected value.")
        self.assertEqual(record['category'], 'sec_struct', 
            "Category from AA1 record do not match the expected value.")
        self.assertEqual(record['pmid'], '557155', 
            "PMID from AA1 record do not match the expected value.")
        self.assertEqual(record['correlation_coefficients'], 
            {'GEIM800105': '0.850', 'ISOY800102': '0.929', 'MAXF760102': '0.891', 
            'PALJ810103': '0.824', 'RACS820111': '0.841', 'ROBB760105': '0.871', 
            'WOEC730101': '-0.806'}, 
            "Correlation Coefficients from AA1 record do not match the expected value.")
        self.assertEqual(record['values'], expected_index_code4_vals, 
            "Values for individual amino acids do not match expected.")
        for aa, val in list(record['values'].items()):
            if (aa == '-'):
                continue
            self.assertIsInstance(val, float, 
                "Values in record values not of correct datatype, expected float, got {}.".format(type(val)))
#5.)
        #testing value error raised when erroneous indices sought from object
        with self.assertRaises(ValueError):
            feature_vals = aaindex1[index_code5]

        with self.assertRaises(ValueError):
            feature_vals = aaindex1[index_code6]

        with self.assertRaises(ValueError):
            feature_vals = aaindex1[index_code7]

    def test_records_dot_notation(self):
            """ Test Case to check that the correct record data and amino acid values 
            are returned for the accessiom number/index codes using the 
            __getitem__ function, accessing the elements using the dot notation created
            from the Map class. """
#1.)
            #initialise test feature codes and their correct amino acid values
            index_code1 = 'KARP850103'
            expected_index_code1_vals = {'-': 0, 'A': 0.892, 'C': 0.925, 'D': 0.932, 
                'E': 0.933, 'F': 0.914, 'G': 0.923, 'H': 0.894, 'I': 0.872, 'K': 1.057, 
                'L': 0.921, 'M': 0.804, 'N': 0.93, 'P': 0.932, 'Q': 0.885, 'R': 0.901, 
                'S': 0.923, 'T': 0.934, 'V': 0.913, 'W': 0.803, 'Y': 0.837}

            index_code2 = 'PALJ810109'
            expected_index_code2_vals = {'-': 0, 'A': 1.15, 'C': 1.03, 'D': 1.0, 'E': 1.37, 
                'F': 0.92, 'G': 0.64, 'H': 0.95, 'I': 0.99, 'K': 1.2, 'L': 1.22, 'M': 1.45, 
                'N': 0.87, 'P': 0.72, 'Q': 1.43, 'R': 1.06, 'S': 0.84, 'T': 0.97, 'V': 0.82, 
                'W': 1.11, 'Y': 0.72}

            index_code3 = 'ROBB760112'
            expected_index_code3_vals = {'-': 0, 'A': -2.5, 'C': -4.7, 'D': 0.0, 'E': -4.4, 
                'F': -4.1, 'G': 4.9, 'H': 1.6, 'I': -3.3, 'K': -0.8, 'L': -2.0, 'M': -4.1, 
                'N': 4.6, 'P': 5.8, 'Q': -0.5, 'R': -1.2, 'S': 2.5, 'T': 1.7, 'V': -3.5, 
                'W': 1.2, 'Y': -0.6}

            index_code4 = 'TANS770103'
            expected_index_code4_vals = {'-': 0, 'A': 0.79, 'C': 1.268, 'D': 0.53, 'E': 0.643, 
                'F': 1.052, 'G': 0.725, 'H': 0.864, 'I': 1.361, 'K': 0.735, 'L': 1.111, 
                'M': 1.092, 'N': 0.832, 'P': 1.249, 'Q': 1.038, 'R': 1.087, 'S': 1.093, 
                'T': 1.214, 'V': 1.428, 'W': 1.114, 'Y': 1.34}

            index_code5 = "ABCDEFGH"
            index_code6 = "123456"
            index_code7 = "blahblahblah"
#2.)
            #get amino acid values for inputted feature/index codes
            record = aaindex1[index_code1]

            self.assertIn('description', list(record.keys()), 
                "description key not found in aaindex record object.")
            self.assertIn('references', list(record.keys()), 
                "reference key not found in aaindex record object.")
            self.assertIn('notes', list(record.keys()), 
                "notes key not found in aaindex record object.")
            self.assertIn('values', list(record.keys()), 
                "values key not found in aaindex record object.")
            self.assertIn('category', list(record.keys()), 
                "category key not found in aaindex record object.")
            self.assertIn('pmid', list(record.keys()), 
                "pmid key not found in aaindex record object.")
            self.assertIn('correlation_coefficients', list(record.keys()), 
                "correlation_coefficients key not found in aaindex record object.")

            self.assertEqual(record.description, "Flexibility parameter for two rigid neighbors (Karplus-Schulz, 1985)",
                "Description from AA1 record do not match the expected value.")
            self.assertEqual(record.notes, '',
                "Notes from AA1 record do not match the expected value.")
            self.assertTrue(record.references.startswith("Karplus, P.A. and Schulz, G.E"),
                "Reference from AA1 record do not match the expected value.")
            self.assertEqual(record.category, 'flexibility', 
                "Category from AA1 record do not match the expected value.")
            self.assertEqual(record.pmid, '', 
                "PMID from AA1 record do not match the expected value.")
            self.assertEqual(record.correlation_coefficients, {}, 
                "Correlation Coefficients from AA1 record do not match the expected value.")
            self.assertEqual(record.values, expected_index_code1_vals, 
                "Values for individual amino acids do not match expected.")
#3.)
            record = aaindex1[index_code2]

            self.assertIn('description', list(record.keys()), 
                "description key not found in aaindex record object.")
            self.assertIn('references', list(record.keys()), 
                "reference key not found in aaindex record object.")
            self.assertIn('notes', list(record.keys()), 
                "notes key not found in aaindex record object.")
            self.assertIn('values', list(record.keys()), 
                "values key not found in aaindex record object.")
            self.assertIn('category', list(record.keys()), 
                "category key not found in aaindex record object.")
            self.assertIn('pmid', list(record.keys()), 
                "pmid key not found in aaindex record object.")
            self.assertIn('correlation_coefficients', list(record.keys()), 
                "correlation_coefficients key not found in aaindex record object.")

            self.assertEqual(record.description, 'Normalized frequency of alpha-helix in alpha/beta class (Palau et al., 1981)',
                "Description from AA1 record do not match the expected value.")
            self.assertEqual(record.notes, '', 
                "Notes from AA1 record do not match the expected value.")
            self.assertTrue(record.references.startswith("Palau, J., Argos, P. and Puigdomenech"), 
                "Reference from AA1 record do not match the expected value.")
            self.assertEqual(record.category, 'sec_struct', 
                "Category from AA1 record do not match the expected value.")
            self.assertEqual(record.pmid, '7118409', 
                "PMID from AA1 record do not match the expected value.")
            self.assertEqual(record.correlation_coefficients, 
                {'AURR980112': '0.817', 'CHOP780201': '0.814', 
                'CRAJ730101': '0.811', 'GEIM800101': '0.816', 'GEIM800104': '0.937', 'ISOY800101': '0.874', 
                'KANM800101': '0.849', 'LEVM780101': '0.898', 'LEVM780104': '0.819', 'MAXF760101': '0.876', 
                'PALJ810102': '0.864', 'PRAM900102': '0.898', 'ROBB760101': '0.805'}, 
                "Correlation Coefficients from AA1 record do not match the expected value.")
            self.assertEqual(record.values, expected_index_code2_vals, 
                "Values for individual amino acids do not match expected.")
#4.)
            record = aaindex1[index_code3]

            self.assertIn('description', list(record.keys()), 
                "description key not found in aaindex record object.")
            self.assertIn('references', list(record.keys()), 
                "reference key not found in aaindex record object.")
            self.assertIn('notes', list(record.keys()), 
                "notes key not found in aaindex record object.")
            self.assertIn('values', list(record.keys()), 
                "values key not found in aaindex record object.")
            self.assertIn('category', list(record.keys()), 
                "category key not found in aaindex record object.")
            self.assertIn('pmid', list(record.keys()), 
                "pmid key not found in aaindex record object.")
            self.assertIn('correlation_coefficients', list(record.keys()), 
                "correlation_coefficients key not found in aaindex record object.")

            self.assertEqual(record.description, 'Information measure for coil (Robson-Suzuki, 1976)',
                "Description from AA1 record do not match the expected value.")
            self.assertEqual(record.notes, '',
                "Notes from AA1 record do not match the expected value.")
            self.assertTrue(record.references.startswith("Robson, B. and Suzuki, E"),
                "Reference from AA1 record do not match the expected value.")
            self.assertEqual(record.category, 'sec_struct', 
                "Category from AA1 record do not match the expected value.")
            self.assertEqual(record.pmid, '1003471', 
                "PMID from AA1 record do not match the expected value.")
            self.assertEqual(record.correlation_coefficients, 
                {'CHOP780211': '0.841', 'ISOY800103': '0.807', 'PALJ810115': '0.885', 
                'QIAN880132': '0.800', 'QIAN880133': '0.814'}, 
                "Correlation Coefficients from AA1 record do not match the expected value.")
            self.assertEqual(record.values, expected_index_code3_vals, 
                "Values for individual amino acids do not match expected.")
#5.)
            record = aaindex1[index_code4]

            self.assertIn('description', list(record.keys()), 
                "description key not found in aaindex record object.")
            self.assertIn('references', list(record.keys()), 
                "reference key not found in aaindex record object.")
            self.assertIn('notes', list(record.keys()), 
                "notes key not found in aaindex record object.")
            self.assertIn('values', list(record.keys()), 
                "values key not found in aaindex record object.")
            self.assertIn('category', list(record.keys()), 
                "category key not found in aaindex record object.")
            self.assertIn('pmid', list(record.keys()), 
                "pmid key not found in aaindex record object.")
            self.assertIn('correlation_coefficients', list(record.keys()), 
                "correlation_coefficients key not found in aaindex record object.")

            self.assertEqual(record.description, 'Normalized frequency of extended structure (Tanaka-Scheraga, 1977)',
                "Description from AA1 record do not match the expected value.")
            self.assertEqual(record.notes, '', 
                "Notes from AA1 record do not match the expected value.")
            self.assertTrue(record.references.startswith("Tanaka, S. and Scheraga, H.A."),
                "Reference from AA1 record do not match the expected value.")
            self.assertEqual(record.category, 'sec_struct', 
                "Category from AA1 record do not match the expected value.")
            self.assertEqual(record.pmid, '557155', 
                "PMID from AA1 record do not match the expected value.")
            self.assertEqual(record.correlation_coefficients, 
                {'GEIM800105': '0.850', 'ISOY800102': '0.929', 'MAXF760102': '0.891', 'PALJ810103': '0.824', 
                'RACS820111': '0.841', 'ROBB760105': '0.871', 'WOEC730101': '-0.806'}, 
                "Correlation Coefficients from AA1 record do not match the expected value.")
            self.assertEqual(record.values, expected_index_code4_vals, 
                "Values for individual amino acids do not match expected.")
#6.)
            #testing value error raised when erroneous indices sought from object
            with self.assertRaises(ValueError):
                feature_vals = aaindex1[index_code5]

            with self.assertRaises(ValueError):
                feature_vals = aaindex1[index_code6]

            with self.assertRaises(ValueError):
                feature_vals = aaindex1[index_code7]

    def test_search(self):
        """ Testing search functionality where 1 or more aaindex1 records can be returned 
        if their description contains user-inputted keywords. """

        description1 = "Gibbs Energy"
        description2 = 'Polarity'
        description3 = 'Hydrophobic'
        description4 = 'pi-helices'
        description5 = 'blahblahblah'
        description6 = 'not a description'
        description7 = 1234
        description8 = False
#1.)
        description1_search = aaindex1.search(description1)
        expected_description1_indices = ['YUTK870101', 'YUTK870102', 'YUTK870103', 'YUTK870104']

        self.assertEqual(len(description1_search), 4, 
            "Expected there to be 4 records returned from search, got {}.".format(len(description1_search)))
        for index, val in description1_search.items():
            self.assertIn(index, expected_description1_indices, 
                "Index record {} expected to be found in search results.".format(index))
            self.assertTrue(description1.lower() in description1_search[index]['description'].lower(), 
                "Expected description to be in record of returned search results: {}".format(description1_search[index]))
#2.)
        description2_search = aaindex1.search(description2)
        expected_description2_indices = ['GRAR740102', 'RADA880108', 'ZIMJ680103']
      
        self.assertEqual(len(description2_search), 3, 
            "Expected there to be 3 records returned from search, got {}.".format(len(description2_search)))
        for index, val in description2_search.items():
            self.assertIn(index, expected_description2_indices,
                "Index record {} expected to be found in search results.".format(index))
            self.assertTrue(description2.lower() in description2_search[index]['description'].lower(),
                "Expected description to be in record of returned search results: {}.".format(description2_search[index]))
#3.)
        description3_search = aaindex1.search(description3)
        expected_description3_indices = ['ARGP820101', 'BLAS910101', 'CASG920101', 'CIDH920101', 'CIDH920102', 'CIDH920103', 
        'CIDH920104', 'CIDH920105', 'COWR900101', 'EISD840101', 'EISD860102', 'EISD860103', 'ENGD860101', 'FASG890101', 
        'FAUJ830101', 'GOLD730101', 'JOND750101', 'JURD980101', 'KIDA850101', 'LEVM760101', 'MANP780101', 'PONP800101', 
        'PONP800102', 'PONP800103', 'PONP800104', 'PONP800105', 'PONP800106', 'PONP930101', 'PRAM900101', 'SWER830101', 
        'WILM950101', 'WILM950102', 'WILM950103', 'WILM950104', 'WOLR790101', 'ZIMJ680101']
 
        self.assertEqual(len(description3_search), 36, 
            "Expected there to be 36 records returned from search, got {}.".format(len(description3_search)))
        for index, val in description3_search.items():
            self.assertIn(index, expected_description3_indices, 
                "Index record {} expected to be found in search results.".format(index))
            self.assertTrue(description3.lower() in description3_search[index]['description'].lower(),
                "Expected description to be in record of returned search results: {}.".format(description3_search[index]))
#4.)
        description4_search = aaindex1.search(description4)
        expected_description4_indices = ['FODM020101']

        self.assertEqual(len(description4_search), 1, 
            "Expected there to be 1 records returned from search, got {}.".format(len(description4_search)))
        for index, val in description4_search.items():
            self.assertIn(index, expected_description4_indices, 
                "Index record {} expected to be found in search results.".format(index))
            self.assertTrue(description4.lower() in description4_search[index]['description'].lower(),
                "Expected description to be in record of returned search results.")
#5.)
        description5_search = aaindex1.search(description5)
        self.assertEqual(len(description5_search), 0, 
            "Expected there to be 0 records returned from search, got {}.".format(len(description5_search)))
#6.)
        description6_search = aaindex1.search(description6)
        self.assertEqual(len(description6_search), 0, 
            "Expected there to be 0 records returned from search, got {}.".format(len(description6_search)))
#7.)    
        with (self.assertRaises(TypeError)):
            description7_search = aaindex1.search(description7)
#8.)
        with (self.assertRaises(TypeError)):
            description8_search = aaindex1.search(description8)

    def test_record_codes(self):
        """ Test Case to check that correct feature/record codes are in the parsed JSON
        of all AAI1 records. Also testing that closely named erroneous records codes are
        not present. """

        #testing actual index codes
        index_code1 = 'VHEG790101'
        index_code2 = 'PONP800107'
        index_code3 = 'OOBM770102'
        index_code4 = 'NADH010101'
#1.)
        #testing index codes that are in the AAI1
        self.assertIn(index_code1, aaindex1.record_codes(), 'Index {} \
            not found in list of available indices.'.format(index_code1))
        self.assertIn(index_code2, aaindex1.record_codes(), 'Index {} \
            not found in list of available indices'.format(index_code2))
        self.assertIn(index_code3, aaindex1.record_codes(), 'Index {} \
            not found in list of available indices'.format(index_code3))
        self.assertIn(index_code4, aaindex1.record_codes(), 'Index {} \
            not found in list of available indices'.format(index_code4))
#2.)
        #testing closely named index codes
        index_code5 = 'VHEG790500'
        index_code6 = 'P0NP800107'
        index_code7 = 'OObm770102'
        index_code8 = 'NADW210101'

        #testing errenous index codes are not in the AAI1
        self.assertNotIn(index_code5, aaindex1.record_codes(), 'Index {} \
            erroneously found in list of available indices.'.format(index_code5))
        self.assertNotIn(index_code6, aaindex1.record_codes(), 'Index {} \
            erroneously found in list of available indices.'.format(index_code6))
        self.assertNotIn(index_code7, aaindex1.record_codes(), 'Index {} \
            erroneously found in list of available indices.'.format(index_code7))
        self.assertNotIn(index_code8, aaindex1.record_codes(), 'Index {} \
            erroneously found in list of available indices.'.format(index_code8))

    def test_aaindex_names(self):
        """ Test Case to check that the correct names/description is returned for each index. """
        
        record_names = aaindex1.record_names()
        record_name1 = record_names[50]
        record_name2 = record_names[140]
        record_name3 = record_names[368]
        record_name4 = record_names[560]
#1.)
        self.assertEqual(len(record_names), aaindex1.num_records(), 
            'Total number of names in the AAI1 database should equal the number of records.')
#2.)
        self.assertEqual(record_name1, 'alpha-NH chemical shifts (Bundi-Wuthrich, 1979)',
            'Incorrect index name returned, got {} instead of {}'.format(record_name1,
                'alpha-NH chemical shifts (Bundi-Wuthrich, 1979).'))
#3.)
        self.assertTrue(record_name2.startswith('Number of full nonbonding'),
            'Incorrect index name returned, got {} instead of {}.'.format(record_name2,
                'Number of full nonbonding'))
#4.)
        self.assertIn('alpha-helix in alpha/beta class', record_name3, 
            'String {} not found in index name: {}'.format(
                'alpha-helix in alpha/beta class', record_name3))
#5.)
        self.assertTrue(record_name4.endswith('Buriability (Zhou-Zhou, 2004)'), 
            'Index does not end with {}, got {}'.format(
                'Buriability (Zhou-Zhou, 2004)', record_name4))

    def test_aaindex_refs(self):
        """ Test Case to check that the correct references are returned for each index. """

        feature1 = 'VELV850101'
        feature2 = 'QIAN880139'
        feature3 = 'NAKH900112'
        feature4 = 'LEVM760103'
        feature5 = "blah_blah"
        feature6 = 123
#1.)
        ref1 = aaindex1[feature1]['references']
        ref2 = aaindex1[feature2]['references']
        ref3 = aaindex1[feature3]['references']
        ref4 = aaindex1[feature4]['references']

        self.assertTrue(ref1.startswith('Veljkovic'), 'Incorrect index reference, \
            got {} instead of {} '.format(ref1,'' ))
        self.assertIn("globular proteins using neural network models", ref2)
        self.assertIn("hydrophobicity of amino acid composition of mitochondrial proteins", ref3)
        self.assertTrue(ref4.endswith('J. Mol. Biol. 104, 59-107 (1976) (Gly missing)'), "{}".format(ref4))
#2.)
        with (self.assertRaises(ValueError)):
            ref5 = aaindex1[feature5]['references']

        with (self.assertRaises(TypeError)):
            ref6 = aaindex1[feature6]['references']

    def test_last_updates(self):
        """ Testing the last updated class attribute which is the latest version of the database. """
        self.assertEqual(aaindex1.last_updated, "February 13, 2017", 
            "Last updated value does not match expected, got {}.".format(aaindex1.last_updated))

    def tearDown(self):
        """ Remove any test data or directories. """
        os.rmdir('test_data')

if __name__ == '__main__':
    #run all unit tests
    unittest.main(verbosity=2)